(***************************************************************************************************
*            F# sample for the usage of TopoART-R (classes TopoART_R and Fast_TopoART_R)           *
*                                 (approximation of the Mandelbrot set)                            *
****************************************************************************************************
*                              Created by Marko Tscherepanow, 9 July 2019                          *
***************************************************************************************************)

// Compile and run from the console: dotnet run --project TopoART-R_sample3.fsproj

// macOS: If libgdiplus cannot be loaded, an installation of mono-libgdiplus via homebrew may be required.
//        Alternatively, the nuget package runtime.osx.10.10-x64.CoreCompat.System.Drawing can be added to the 
//        project, for instance, by entering the following command in the project directory: 
//        dotnet add package runtime.osx.10.10-x64.CoreCompat.System.Drawing 
// openSUSE, Raspberry Pi OS, Ubuntu: If libgdiplus cannot be loaded, its installation using the package manager
//        may be required.

/// <summary>
/// Pixel-wise regression analysis of an image using TopoART-R. [F#]
/// <para>
/// This sample trains a TopoART-R network with a colour image depicting the Mandelbrot set. The 
/// pixel coordinates are used as independent variables and the corresponding colour values as 
/// dependent variables. The accuracy of the regression function can be controlled by means of the 
/// vigilance parameter <c>rho_a</c>.
/// </para>
/// <para>The training image <c>Mandelbrot_reference.png</c> and the predicted image 
/// <c>TopoART-R_Mandelbrot_regression.png</c> are written into the subfolder 
/// <c>results/regression</c>.
/// </para>
/// </summary>
module LibTopoART_samples.TopoART_R_sample3

open LibTopoART
open System
open System.Drawing
open System.Drawing.Imaging
open System.IO
open System.Reflection

type Selector = USE_TopoART_R | USE_Fast_TopoART_R

// Choose TopoART-R implementation (see enum Selector above)
let sel = USE_Fast_TopoART_R

// Destination directory for the regression results
let ResultPath = "../../../../../results/regression/"
let NetworkPath = "../../../../../results/networks/"

// Input and output dimensions
let i_len = 2; // Number of independent variables (input dimensions)
let d_len = 3; // Number of dependent variables (output dimensions)

// TopoART-R parameters
let rho_a = 0.95m    // Most relevant parameter (decrease and/or increase to visualise effects)
let beta_sbm = 0.5m  // Default
let phi = 1L         // Default (increase to enable noise reduction)
let tau = 100L       // Default

// Application parameters
let img_width = 800
let img_height = 600
let max_iter = 1023

(*--------------------------------------------------------------------------------------------------
-                                           Helper functions                                       -
--------------------------------------------------------------------------------------------------*)

let ColorFromIteration iter =
    if iter = max_iter then
        Color.FromArgb(0, 0, 0, 0)
    else
        let b_tmp = float iter / float max_iter * 100.0
        let b = int (b_tmp / ( 1.0 + b_tmp ) * 255.0)
        let g_tmp = float iter / float max_iter * 25.0
        let g = int (g_tmp / ( 1.0 + g_tmp ) * 255.0)
        let r_tmp = float iter / float max_iter * 5.0
        let r = int ((1.0 - (r_tmp / ( 1.0 + r_tmp ))) * 31.0)
        Color.FromArgb(0, r, g, b)

let ColorFromPrediction (pred : decimal[]) =
    let r = Math.Round(pred.[0] * 255.0m)
    let g = Math.Round(pred.[1] * 255.0m)
    let b = Math.Round(pred.[2] * 255.0m)
    Color.FromArgb(0, int r, int g, int b)

let GetPixel x y =
    let x0 = float x / float img_width * 3.0 - 2.0
    let y0 = float y / float img_height * 2.0 - 1.0

    let mutable x = 0.0
    let mutable y = 0.0
    let mutable i = 0

    while x * x + y * y <= 4.0 && i < max_iter do 
        let tmp = x * x - y * y + x0
        y <- 2.0 * x * y + y0
        x <- tmp
        i <- i + 1
    i

(*--------------------------------------------------------------------------------------------------
-                                             Main program                                         -
--------------------------------------------------------------------------------------------------*)

[<EntryPoint>]
let main args =

    // Set working directory to assembly directory
    let cb = new Uri(Assembly.GetEntryAssembly().Location)
    Directory.SetCurrentDirectory(Path.GetDirectoryName(cb.LocalPath))

    let sample_num = img_width * img_height
    let i = Array.create sample_num <| Array.create i_len 0.0m
    let d = Array.create sample_num <| Array.create d_len 0.0m
    
    printf "Generate training data\n"

    // Generate data (used for training and testing)
    let mutable sample_index = 0
    let ref_bm = new Bitmap(img_width, img_height, PixelFormat.Format24bppRgb)
    for y = 0 to img_height - 1 do
        for x = 0 to img_width - 1 do
            let iter = GetPixel x y
            let c = ColorFromIteration iter
            ref_bm.SetPixel(x, y, c)
            i.[sample_index] <- [| (decimal x) / (decimal (img_width - 1)); (decimal y) / (decimal (img_height - 1)) |]
            d.[sample_index] <- [| decimal c.R / 255.0m; decimal c.G / 255.0m; decimal c.B / 255.0m |]
            sample_index <- sample_index + 1

    // Save reference image
    ref_bm.Save(ResultPath + "Mandelbrot_reference.png") 

    // Create TopoART-R network
    let tar = 
        if sel = USE_TopoART_R then 
            new TopoART_R(int64 i_len, int64 d_len, 2L, rho_a) :> ITopoART_R
        else
            new Fast_TopoART_R(int64 i_len, int64 d_len, 2L, rho_a) :> ITopoART_R

    let FilePrefix = if sel = USE_TopoART_R then "TopoART-R" else "Fast_TopoART-R"
    let FileSuffix = if sel = USE_TopoART_R then "tar" else "ftar"

    tar.Beta_sbm <- beta_sbm
    tar.Tau <- tau
    tar.Phi <- phi
    
    // Start time measuring
    let training_start = DateTime.Now

    // Train
    for train_index = 0 to sample_num - 1 do
        tar.Learn(i.[train_index], d.[train_index])
        if train_index % 10000 = 0 then printf "."
    printf "\n"

    // Stop time measuring
    let training_end = DateTime.Now

    // Output the required time
    let training_time = training_end - training_start
    printf "Time for training: %s\n" <| training_time.ToString()

    let result_bm = new Bitmap(img_width, img_height, PixelFormat.Format24bppRgb)

    // Start time measuring
    let prediction_start = DateTime.Now

    // Predict
    for predict_index = 0 to sample_num - 1 do
        let x = predict_index % img_width
        let y = predict_index / img_width
        let prediction = tar.Predict(i.[predict_index])
        let c = ColorFromPrediction prediction
        result_bm.SetPixel(x, y, c)
        if predict_index % 10000 = 0 then printf "."
    printf "\n"

    // Stop time measuring
    let prediction_end = DateTime.Now

    // Output the required time
    let prediction_time = prediction_end - prediction_start
    printf "Time for prediction: %s\n" <| prediction_time.ToString()

    // Save reference image
    result_bm.Save(ResultPath + FilePrefix + "_Mandelbrot_regression.png") 

    printf "Save network\n"

    // Save network in human readable form
    // tar.SaveText(NetworkPath + FilePrefix + "_Mandelbrot_regression.txt");

    // Save network in binary form
    tar.Save(NetworkPath + FilePrefix + "_Mandelbrot_regression." + FileSuffix);

    0
